# pcbnew_easy.py
import math
import cmath

from kicad import pcbnew_bare as pcbnew

print('pcbnew_easy is not supported in KiCad >= 5.99\nUse the universal interface with "from kicad import pcbnew"')

# helper functions
def _from_mm(val):
    """Convert mm to internal units"""
    return pcbnew.FromMM(float(val))

def _point_mm(x, y):
    """Convert coordinate in mm to internal coordinate"""
    return pcbnew.wxPointMM(float(x), float(y))

def _size_mm(x, y):
    """Convert size in mm to internal size"""
    return pcbnew.wxSizeMM(float(x), float(y))

def rotate(coord, angle):
    """Rotate coordinate around (0, 0)"""
    coord = (coord[0]+coord[1]*1j) * cmath.exp(math.radians(angle)*1j)
    return (coord.real, coord.imag)

# dicts for converting layer name to id, used by _get_layer
layer_dict = {pcbnew.BOARD_GetStandardLayerName(n):n for n in range(pcbnew.PCB_LAYER_ID_COUNT)}
layer_names = {s:n for n, s in layer_dict.items()}

def _get_layer(s):
    """Get layer id from layer name

    If it is already an int just return it.
    """
    return layer_dict[s]

def _to_LayerSet(layers):
    """Create LayerSet used for defining pad layers"""
    bitset = 0
    for l in layers:
        bitset |= 1 << _get_layer(l)
    hexset = '{0:013x}'.format(bitset)
    lset = pcbnew.LSET()
    lset.ParseHex(hexset, len(hexset))
    return lset

def _from_LayerSet(layerset):
    mask = [c for c in layerset.FmtBin() if c in ('0','1')]
    mask.reverse()
    ids = [i for i, c in enumerate(mask) if c == '1']
    return tuple(layer_names[i] for i in ids)



class Module(object):
    def __init__(self, module):
        """Convenience wrapper for pcbnew Module"""
        self._module = module

    @property
    def position(self):
        return pcbnew.ToMM(self._module.GetPosition())
    @position.setter
    def position(self, position):
        self._module.SetPosition(_point_mm(position[0], position[1]))

    @property
    def reference(self):
        return self._module.GetReference()
    @reference.setter
    def reference(self, value):
        self._module.SetReference(value)

    @property
    def value(self):
        return self._module.GetValue()
    @value.setter
    def value(self, value):
        self._module.SetValue(value)

    @property
    def pads(self):
        p = self._module.Pads().begin()
        while not p == None:
            yield Pad(p)
            p = p.Next()

    def flip(self, center=None):
        if center==None:
            center = self.position
        self._module.Flip(_point_mm(center[0], center[1]))

    def add_line(self, start, end, layer='F.SilkS', width=0.15):
        """Create a graphic line on the module"""
        a = pcbnew.EDGE_MODULE(self._module)
        a.SetShape(pcbnew.S_SEGMENT)
        a.SetStart(_point_mm(start[0], start[1]))
        a.SetEnd(_point_mm(end[0], end[1]))
        a.SetLayer(_get_layer(layer))
        a.SetWidth(_from_mm(width))
        a.SetLocalCoord()
        self._module.Add(a)
        return a

    def add_polyline(self, coords, layer='F.SilkS', width=0.15):
        """Create a graphic polyline on the module"""
        for n in range(len(coords)-1):
            self.add_line(coords[n], coords[n+1], layer=layer, width=width)

    def add_circle(self, center, radius, layer='F.SilkS', width=0.15):
        """Create a graphic circle on the module"""
        a = pcbnew.EDGE_MODULE(self._module)
        a.SetShape(pcbnew.S_CIRCLE)
        a.SetCenter(_point_mm(center[0], center[1]))
        start_coord = _point_mm(center[0], center[1]+radius)
        a.SetArcStart(start_coord)
        a.SetLayer(_get_layer(layer))
        a.SetWidth(_from_mm(width))
        a.SetLocalCoord()
        self._module.Add(a)
        return a

    def add_arc(self, center, radius, start_angle, stop_angle, layer='F.SilkS', width=0.15):
        """Create a graphic arc on the module"""
        start_coord = radius * cmath.exp(math.radians(start_angle-90)*1j)
        start_coord = _point_mm(start_coord.real, start_coord.imag)
        angle = stop_angle - start_angle
        a = pcbnew.EDGE_MODULE(self._module)
        a.SetShape(pcbnew.S_ARC)
        a.SetCenter(_point_mm(center[0], center[1]))
        a.SetArcStart(start_coord)
        a.SetAngle(angle*10)
        a.SetLayer(_get_layer(layer))
        a.SetWidth(_from_mm(width))
        a.SetLocalCoord()
        self._module.Add(a)
        return a

    def add_pad(self, position, size, name='', pad_type='standard', shape='circle',
                drill=1.0, layers=None):
        """Create a pad on the module

        Args:
            position: pad position in mm
            size: pad size in mm, value if shape == 'circle', tuple otherwise
            name: pad name/number
            pad_type: One of 'standard', 'smd', 'conn', 'hole_not_plated'
            shape: One of 'circle', 'rect', 'oval', 'trapezoid'
            drill: drill size in mm, single value for round hole, or tuple for oblong hole.
            layers: None for default, or a list of layer definitions (for example: ['F.Cu', 'F.Mask'])
        """
        pad = Pad(pcbnew.D_PAD(self._module))

        pad.type = pad_type
        pad.shape = shape
        pad.size = size
        pad.name = name
        pad.position = position
        pad.layers = layers

        self._module.Add(pad._pad)
        return pad

    def save(self, library_path):
        """Save footprint in given library

        library_path should end with .pretty
        """
        self._module.SetFPID(pcbnew.FPID(self._module.GetReference()))

        io = pcbnew.PCB_IO()

        try:
            io.FootprintLibCreate(library_path)
        except IOError:
            pass # we try to create, but may be it exists already

        io.FootprintSave(library_path, self._module)

class Pad(object):
    """Convenience wrapper for pcbnew Pad"""
    _pad_types = {'standard':pcbnew.PAD_STANDARD,
                  'smd':pcbnew.PAD_SMD,
                  'conn':pcbnew.PAD_CONN,
                  'hole_not_plated':pcbnew.PAD_HOLE_NOT_PLATED}
    _pad_types_id = {s:n for n, s in _pad_types.items()}
    _shapes = {'circle':pcbnew.PAD_CIRCLE,
               'rect':pcbnew.PAD_RECT,
               'oval':pcbnew.PAD_OVAL,
               'trapezoid':pcbnew.PAD_TRAPEZOID}
    _shapes_id = {s:n for n, s in _shapes.items()}

    def __init__(self, pad):
        self._pad = pad

    @property
    def position(self):
        return pcbnew.ToMM(self._pad.GetPosition())
    @position.setter
    def position(self, position):
        self._pad.SetPosition(_point_mm(position[0], position[1]))
        self._pad.SetLocalCoord()

    @property
    def name(self):
        return self._pad.GetPadName()
    @name.setter
    def name(self, value):
        self._pad.SetPadName(value)

    @property
    def type(self):
        return self._pad_types_id[self._pad.GetAttribute()]
    @type.setter
    def type(self, value):
        self._pad.SetAttribute(self._pad_types[value])

    @property
    def shape(self):
        return self._shapes_id[self._pad.GetShape()]
    @shape.setter
    def shape(self, value):
        self._pad.SetShape(self._shapes[value])

    @property
    def size(self):
        if self.shape == 'circle':
            return pcbnew.ToMM(self._pad.GetSize())[0]
        else:
            return pcbnew.ToMM(self._pad.GetSize())
    @size.setter
    def size(self, size):
        if self.shape == 'circle':
            self._pad.SetSize(_size_mm(size, size))
        else:
            self._pad.SetSize(_size_mm(size[0], size[1]))

    @property
    def orientation(self):
        return self._pad.GetOrientation()/10.0
    @orientation.setter
    def orientation(self, value):
        self._pad.SetOrientation(value*10)

    @property
    def layers(self):
        return _from_LayerSet(self._pad.GetLayerSet())
    @layers.setter
    def layers(self, value):
        if value == None: # if None set default layers
            default_masks = {'standard':self._pad.StandardMask(),
                             'smd':self._pad.SMDMask(),
                             'conn':self._pad.ConnSMDMask(),
                             'hole_not_plated':self._pad.UnplatedHoleMask()}
            self._pad.SetLayerSet(default_masks[self.type])
        else:
            self._pad.SetLayerSet(_to_LayerSet(value))

    @property
    def drill(self):
        if self.type in ('standard', 'hole_not_plated'):
            if self._pad.GetDrillShape() == pcbnew.PAD_DRILL_OBLONG:
                return pcbnew.ToMM(self._pad.GetDrillSize())
            else:
                return pcbnew.ToMM(self._pad.GetDrillSize())[0]
        else:
            return None
    @drill.setter
    def drill(self, drill):
        if self.type in ('standard', 'hole_not_plated'):
            if hasattr(drill, '__getitem__'):
                self._pad.SetDrillShape(pcbnew.PAD_DRILL_OBLONG)
                self._pad.SetDrillSize(_size_mm(drill[0], drill[1]))
            else:
                self._pad.SetDrillSize(_size_mm(drill, drill))
        else:
            pass

def get_board():
    """Get the current board"""
    return Board(pcbnew.GetBoard())


# Usage example and test
def test():
    """Make an example board

    Run in the pcbnew scripting console:
      import pcbnew_easy
      pcbnew_easy.test()
    """
    # get current board
    pcb = get_board()

    # create test module
    m = pcb.create_module('test')
    m.add_arc(center=(0, 0), radius=8, start_angle=-90, stop_angle=90, width=0.2)
    m.add_line(start=(-8, 0), end=(8, 0), width=0.2)
    m.add_pad(position=(-4, -3), size=2, drill=1)
    m.add_pad(position=(4, -3), size=2, drill=1, layers=['B.Cu', 'F.Cu'])
    for n, x in enumerate([-1, -.5, 0, .5, 1]):
        m.add_pad(position=(x, -4), size=(0.25, 1.2), name=n, pad_type='smd', shape='rect')

    # move module to right location
    m.position = (30, 30)

    # add test track with via
    track1 = [(30, 26), (30, 50), (60, 80)]
    track2 = [(60, 80), (80, 80)]
    pcb.add_track(track1, layer='F.Cu', width=0.25)
    pcb.add_track_via(track1[-1])
    pcb.add_track(track2, layer='B.Cu')

    # add board edge
    ul = (20, 20)
    pcb_size = (100, 80)
    edge = [ul,
            (ul[0], ul[1]+pcb_size[1]),
            (ul[0]+pcb_size[0], ul[1]+pcb_size[1]),
            (ul[0]+pcb_size[0], ul[1]),
            ul]
    pcb.add_polyline(edge, layer='Edge.Cuts')

    print('List all pads of all modules')
    for module in pcb.modules:
        print(module.reference)
        for pad in m.pads:
            print('  {0.name}: {0.shape}'.format(pad))



from kicad import pcbnew_bare as pcbnew

import kicad
from kicad.pcbnew import layer as pcbnew_layer
from kicad.point import Point
from kicad import units, SWIGtype, SWIG_version
from kicad.pcbnew.item import HasConnection, HasLayerStrImpl, Selectable, BoardItem
from kicad.pcbnew.layer import LayerSet

class KeepoutAllowance(object):
    """ Gives key-value interface of the form
            my_zone.is_keepout = True
            my_zone.allow['tracks'] = False
    """
    def __init__(self, zone):
        self._zone = zone

    def __getitem__(self, item):
        # if not self._zone.is_keepout:
        #     print('Warning: '
        #         'Keepout settings do not apply to fill zones.'
        #         ' Call "my_zone.is_keepout = True" first.'
        #     )
        if item == 'tracks':
            return not self._zone._obj.GetDoNotAllowTracks()
        elif item == 'pour':
            return not self._zone._obj.GetDoNotAllowCopperPour()
        elif item == 'vias':
            return not self._zone._obj.GetDoNotAllowVias()
        else:
            raise KeyError(
                'Invalid zone keepout allowance type: {}. '
                'Allowed types are "tracks, pour, vias"'.format(item)
            )

    def __setitem__(self, item, value):
        # if not self._zone.is_keepout:
        #     print('Warning: '
        #         'Keepout settings do not apply to fill zones.'
        #         ' Call "my_zone.is_keepout = True" first.'
        #     )
        if item == 'tracks':
            self._zone._obj.SetDoNotAllowTracks(not bool(value))
        elif item == 'pour':
            self._zone._obj.SetDoNotAllowCopperPour(not bool(value))
        elif item == 'vias':
            self._zone._obj.SetDoNotAllowVias(not bool(value))
        else:
            raise KeyError(
                'Invalid zone keepout allowance type: {}. '
                'Allowed types are "tracks, pour, vias"'.format(item)
            )

    def __str__(self):
        return str({k: self[k] for k in ['tracks', 'pour', 'vias']})

    def __repr__(self):
        return type(self).__name__ + str(self)


class Zone(HasConnection, HasLayerStrImpl, Selectable, BoardItem):
    def __init__(self, layer='F.Cu', board=None):
        self._obj = SWIGtype.Zone(board and board.native_obj)
        self.layer = layer
        raise NotImplementedError('Constructor not supported yet')

    @staticmethod
    def wrap(instance):
        """Wraps a C++ api TRACK object, and returns a `Track`."""
        return kicad.new(Zone, instance)

    @property
    def clearance(self):
        if SWIG_version >= 7:
            native = self._obj.GetLocalClearance()
        else:
            native = self._obj.GetClearance()
        return float(native) / units.DEFAULT_UNIT_IUS

    @clearance.setter
    def clearance(self, value):
        if SWIG_version >= 7:
            self._obj.SetLocalClearance(int(value * units.DEFAULT_UNIT_IUS))
        else:
            self._obj.SetClearance(int(value * units.DEFAULT_UNIT_IUS))
            self._obj.SetZoneClearance(int(value * units.DEFAULT_UNIT_IUS))

    @property
    def min_width(self):
        return float(self._obj.GetMinThickness()) / units.DEFAULT_UNIT_IUS

    @min_width.setter
    def min_width(self, value):
        self._obj.SetMinThickness(int(value * units.DEFAULT_UNIT_IUS))

    @property
    def is_keepout(self):
        if SWIG_version >= 6:
            return bool(self._obj.GetIsRuleArea())
        else:
            return bool(self._obj.GetIsKeepout())

    @is_keepout.setter
    def is_keepout(self, value):
        if SWIG_version >= 6:
            self._obj.SetIsRuleArea(bool(value))
        else:
            self._obj.SetIsKeepout(bool(value))

    @property
    def allow(self):
        return KeepoutAllowance(self)

    def delete(self):
        self._obj.DeleteStructure()

    @property
    def layerset(self):
        ''' For zones with multiple layers
            Changing this layerset will not propagate back to this zone
            until you set layerset again. Common pattern:
                zone.layerset = zone.layerset.add_layer('F.Cu')
        '''
        from kicad.pcbnew.board import Board
        layers_native = self._obj.GetLayerSet()
        lset = LayerSet.wrap(layers_native)
        lset._board = Board.wrap(self._obj.GetBoard())
        return lset

    @layerset.setter
    def layerset(self, new_lset):
        self._obj.SetLayerSet(new_lset._obj)

# GetCornerSmoothingType
# GetDefaultHatchPitch

# GetThermalReliefCopperBridge
# GetThermalReliefGap

# Outline
# RawPolysList



from kicad import pcbnew_bare as pcbnew

import kicad
from kicad.pcbnew import layer as pcbnew_layer
from kicad.point import Point
from kicad import units, SWIGtype
from kicad.pcbnew.item import HasConnection, HasLayerStrImpl, Selectable, BoardItem

class Track(HasConnection, HasLayerStrImpl, Selectable, BoardItem):
    def __init__(self, width, start, end, layer='F.Cu', board=None):
        self._obj = SWIGtype.Track(board and board.native_obj)
        self._obj.SetWidth(int(width * units.DEFAULT_UNIT_IUS))
        self.layer = layer
        self._obj.SetStart(Point.native_from(start))
        self._obj.SetEnd(Point.native_from(end))

    @staticmethod
    def wrap(instance):
        """Wraps a C++ api TRACK object, and returns a `Track`."""
        return kicad.new(Track, instance)

    @property
    def width(self):
        return float(self._obj.GetWidth()) / units.DEFAULT_UNIT_IUS

    @width.setter
    def width(self, value):
        self._obj.SetWidth(int(value * units.DEFAULT_UNIT_IUS))

    @property
    def start(self):
        return Point.wrap(self._obj.GetStart())

    @start.setter
    def start(self, value):
        self._obj.SetStart(Point.native_from(value))

    @property
    def end(self):
        return Point.wrap(self._obj.GetEnd())

    @end.setter
    def end(self, value):
        self._obj.SetEnd(Point.native_from(value))

    def delete(self):
        self._obj.DeleteStructure()



from functools import wraps

class NoDefaultUnits(Exception):

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


def notify(*args):
    ''' Show text in a popup window while in the GUI.
        Arguments act the same as print(*args)
        It is not the best debugging tool ever created, but
        it is handy for debugging action plugins
    '''
    text = ' '.join(str(arg) for arg in args)
    try:
        import wx
    except ImportError:
        print(text)
        return
    try:
        dialog = wx.MessageDialog(None, text, 'kicad-python debug output', wx.OK)
        sg = dialog.ShowModal()
        return sg
    except:
        print(text)


deprecate_warn_fun = notify  # print is sometimes good
def deprecate_member(old, new, deadline='v0.5.0'):
    def regular_decorator(klass):
        def auto_warn(fun):
            from_str = klass.__name__ + '.' + old
            to_str = klass.__name__ + '.' + new
            header = 'Deprecation warning (deadline {}): '.format(deadline)
            map_str = '{} -> {}'.format(from_str, to_str)
            @wraps(fun)
            def warner(*args, **kwargs):
                deprecate_warn_fun(header + map_str)
                return fun(*args, **kwargs)
            return warner

        new_meth = getattr(klass, new)
        if isinstance(new_meth, property):
            aug_meth = property(
                auto_warn(new_meth.fget),
                auto_warn(new_meth.fset),
                auto_warn(new_meth.fdel)
            )
        else:
            aug_meth = auto_warn(new_meth)
        setattr(klass, old, aug_meth)
        return klass
    return regular_decorator



import unittest

from kicad import *


class TestKicadUnits(unittest.TestCase):

    def test_converters(self):
        self.assertEqual(inch_to_mm(1), 25.4)
        self.assertEqual(mm_to_inch(25.4), 1.0)

    def test_converters_seq(self):
        self.assertEqual(inch_to_mm([1, 2]), [25.4, 50.8])
        self.assertEqual(mm_to_inch([25.4, 50.8]), [1.0, 2.0])

    def test_base_unit_tuple_mm(self):
        bunit = Point(1, 2)
        self.assertEqual(bunit.mm, (1.0, 2.0))

    def test_base_unit_tuple_inch(self):
        bunit = Point(1 * inch, 2 * inch)
        self.assertEqual(bunit.inch, (1.0, 2.0))

    def test_base_unit_tuple_mil(self):
        bunit = Point(1 * mil, 2 * mil)
        self.assertEqual(bunit.mil, (1.0, 2.0))

    def test_base_unit_tuple_nm(self):
        bunit = Point(1 * nm, 2 * nm)
        self.assertEqual(bunit.nm, (1, 2))



from kicad import pcbnew_bare
import kicad
from kicad import SWIGtype

from kicad.units import *


class Size(BaseUnitTuple):

    def __init__(self, width, height):
        self._class = Size
        self._obj = SWIGtype.Size(int(width * DEFAULT_UNIT_IUS),
                                  int(height * DEFAULT_UNIT_IUS))

    @staticmethod
    def wrap(instance):
        """Takes a wxSize instance and returns a Size class."""
        wrapped_size = kicad.new(Size, instance)
        wrapped_size._class = Size
        return wrapped_size

    @property
    def native_obj(self):
        return self._obj

    @staticmethod
    def build_from(t):
        return Size._tuple_to_class(t, Size)

    @staticmethod
    def native_from(t):
        return Size._tuple_to_class(t, Size).native_obj

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "Size(%g, %g)" % self.mm

    def scale(self, x, y):
        """Scale this size by x horizontally, and y vertically."""
        scaled = self.scaled(x, y)
        self.x = scaled.x
        self.y = scaled.y

    def scaled(self, x, y):
        """Return a new scaled point, scaling by x and y."""
        scaled = self.Scale(x, y)
        return Size(scaled.x, scaled.y)

    @property
    def width(self):
        """Return the width of the size."""
        return self.x

    @width.setter
    def width(self, value):
        """Set the width of the size."""
        self.x = value

    @property
    def height(self):
        """Return the height of the size."""
        return self.y

    @height.setter
    def height(self, value):
        """Set the height of the size."""
        self.y = value



import unittest

from kicad import *


class TestKicadSize(unittest.TestCase):

    def test_size_properties(self):
        size = Size(1, 2)
        self.assertEqual(1, size.width)
        self.assertEqual(2, size.height)

    def test_addition(self):
        s1 = Size(1, 2)
        s2 = Size(2, 3)
        self.assertEqual((3, 5), s1 + s2)

    def test_substraction(self):
        s1 = Size(3, 4)
        s2 = Size(1, 1)
        self.assertEqual((2, 3), s1 - s2)



import unittest

from kicad import *
from kicad.pcbnew.drawing import *
from kicad import obj


class TestPcbnewDrawing(unittest.TestCase):
    def test_segment_from_native(self):
        native_segment = Segment((0, 0), (1, 1))._obj
        new_segment = obj.wrap(native_segment)
        self.assertEqual(Segment, type(new_segment))

    def test_circle_from_native(self):
        native_circle = Circle((0, 0), 10.0)._obj
        new_circle = obj.wrap(native_circle)
        self.assertEqual(Circle, type(new_circle))

    def test_arc_from_native(self):
        native_arc = Arc((0, 0), 10.0, -90, 90)._obj
        new_arc = obj.wrap(native_arc)
        self.assertEqual(Arc, type(new_arc))
        self.assertEqual(new_arc.angle, 180)
        new_arc.angle = 40  # This will be rounded within 1e-3
        self.assertTrue(abs(new_arc.angle - 40) < 0.001)



from math import radians, degrees
from kicad.point import Point
from kicad.exceptions import deprecate_member
import kicad.pcbnew.layer as pcbnew_layer

class BoardItem(object):
    _obj = None

    @property
    def native_obj(self):
        return self._obj

    @property
    def board(self):
        from kicad.pcbnew.board import Board
        brd_native = self._obj.GetBoard()
        if brd_native:
            return Board(brd_native)
        else:
            return None


class HasPosition(object):
    """Board items that has valid position property should inherit
    this."""

    def __init__(self):
        raise NotImplementedError("This is an abstract class!")

    @property
    def position(self):
        return Point.wrap(self._obj.GetPosition())

    @position.setter
    def position(self, value):
        self._obj.SetPosition(Point.native_from(value))

    @property
    def x(self):
        return self.position.x

    @x.setter
    def x(self, value):
        self.position = (value, self.y)

    @property
    def y(self):
        return self.position.y

    @y.setter
    def y(self, value):
        self.position = (self.x, value)


class HasRotation(object):
    """Board items that has rotation property should inherit this."""
    def __init__(self):
        raise NotImplementedError("This is an abstract class!")

    @property
    def rotation(self):
        """Rotation of the item in radians."""
        return radians(self._obj.GetOrientation() / 10.)

    @rotation.setter
    def rotation(self, value):
        self._obj.SetOrientation(degrees(value) * 10.)


class HasLayerEnumImpl(object):
    """Board items that has layer should inherit this."""
    def __init__(self):
        raise NotImplementedError("This is an abstract class!")

    @property
    def layer(self):
        return pcbnew_layer.Layer(self._obj.GetLayer())

    @layer.setter
    def layer(self, value):
        self._obj.SetLayer(value.value)


class HasLayer(HasLayerEnumImpl):
    _has_warned = False

    def print_warning(self):
        if not self._has_warned:
            print('\nDeprecation warning (HasLayer): Use either HasLayerEnumImpl or HasLayerStrImpl.'
                  '\nDefault will change from Enum to Str in the future.')
            self._has_warned = True

    @property
    def layer(self):
        self.print_warning()
        return pcbnew_layer.Layer(self._obj.GetLayer())

    @layer.setter
    def layer(self, value):
        self.print_warning()
        self._obj.SetLayer(value.value)


class HasLayerStrImpl(object):
    """ Board items that has layer outside of standard layers should inherit this.
        String implementation can sometimes be more intuitive, and accomodates custom layer names.
        If the layer is not present, it will be caught at runtime, rather than disallowed.
    """
    def __init__(self):
        raise NotImplementedError("This is an abstract class!")

    @property
    def layer(self):
        layid = self._obj.GetLayer()
        try:
            brd = self.board
        except AttributeError:
            from kicad.pcbnew.board import Board
            native = self._obj.GetBoard()
            brd = Board(native) if native else None
        return pcbnew_layer.get_board_layer_name(brd, layid)

    @layer.setter
    def layer(self, value):
        try:
            brd = self.board
        except AttributeError:
            from kicad.pcbnew.board import Board
            native = self._obj.GetBoard()
            brd = Board(native) if native else None
        layid = pcbnew_layer.get_board_layer_id(brd, value)
        self._obj.SetLayer(layid)


@deprecate_member('netName', 'net_name')
@deprecate_member('netCode', 'net_code')
class HasConnection(object):
    """All BOARD_CONNECTED_ITEMs should inherit this."""
    def __init__(self):
        raise NotImplementedError("This is an abstract class!")

    @property
    def net_name(self):
        return self._obj.GetNetname()

    @net_name.setter
    def net_name(self, value):
        """ Takes a name and attempts to look it up based on the containing board """
        if not self._obj:
            raise TypeError("Cannot set net_name without a containing Board.")
        try:
            new_code = self._obj.GetBoard().GetNetcodeFromNetname(value)
        except IndexError:
            raise KeyError("Net name '{}' not found in board nets.".format(value))
        self._obj.SetNetCode(new_code)

    @property
    def net_code(self):
        return self._obj.GetNetCode()

    @net_code.setter
    def net_code(self, value):
        self._obj.SetNetCode(value)


class Selectable(object):
    """ This influences the main window. Make sure to pcbnew.Refresh() to see it """
    def __init__(self):
        raise NotImplementedError("This is an abstract class!")

    @property
    def is_selected(self):
        return bool(self._obj.IsSelected())

    def select(self, value=True):
        """ Selecting changes the appearance and also plays a role in determining
            what will be the subject of a subsequent command (delete, move to layer, etc.)
        """
        if value:
            self._obj.SetSelected()
        else:
            self._obj.ClearSelected()

    def deselect(self):
        self.select(False)

    def brighten(self, value=True):
        """ Brightening gives a bright green appearance """
        if value:
            self._obj.SetBrightened()
        else:
            self._obj.ClearBrightened()


class HasWidth(object):
    @property
    def width(self):
        return float(self._obj.GetWidth()) / units.DEFAULT_UNIT_IUS

    @width.setter
    def width(self, value):
        self._obj.SetWidth(int(value * units.DEFAULT_UNIT_IUS))



from kicad import pcbnew_bare

import kicad
from kicad import SWIGtype
from kicad.pcbnew import board
from kicad.pcbnew import drawing
from kicad.pcbnew import module


_WRAPPERS = {pcbnew_bare.BOARD: board.Board,
             SWIGtype.Shape: drawing.Drawing,
             SWIGtype.Footprint: module.Module,
             SWIGtype.Point: kicad.Point,
             SWIGtype.Size: kicad.Size}


def wrap(instance):
    """Returns a python wrapped object from a swig/C++ one."""
    if type(instance) in _WRAPPERS:
        return _WRAPPERS[type(instance)].wrap(instance)

    raise ValueError("Class with no wrapper: %s" % type(instance))



from kicad import pcbnew_bare as pcbnew

import kicad
from kicad.exceptions import deprecate_member
from kicad import Point, Size, DEFAULT_UNIT_IUS, SWIGtype, SWIG_version
from kicad.pcbnew.item import HasPosition, HasRotation, HasLayerEnumImpl, Selectable, HasLayerStrImpl, BoardItem
from kicad.pcbnew.pad import Pad


class ModuleLabel(HasPosition, HasRotation, HasLayerStrImpl, Selectable):
    """wrapper for `TEXTE_MODULE`"""
    def __init__(self, mod, text=None, layer=None):
        self._obj = SWIGtype.FpText(mod.native_obj)
        mod.native_obj.Add(self._obj)
        if text:
            self.text = text
        if layer:
            self.layer = layer

    @property
    def text(self):
        return self._obj.GetText()

    @text.setter
    def text(self, value):
        return self._obj.SetText(value)

    @property
    def visible(self):
        raise ValueError('ModuleLabel.visible is write only.')

    @visible.setter
    def visible(self, value):
        return self._obj.SetVisible(value)

    @property
    def thickness(self):
        return float(self._obj.GetThickness()) / DEFAULT_UNIT_IUS

    @thickness.setter
    def thickness(self, value):
        return self._obj.SetThickness(int(value * DEFAULT_UNIT_IUS))

    @property
    def size(self):
        return Size.wrap(self._obj.GetTextSize())

    @size.setter
    def size(self, value):
        if isinstance(value, tuple):
            if not isinstance(value, Size):
                value = Size(value[0], value[1])
            self._obj.SetTextSize(value.native_obj)

        else: # value is a single number/integer
            self._obj.SetTextSize(Size(value, value).native_obj)

    @staticmethod
    def wrap(instance):
        if type(instance) is SWIGtype.FpText:
            return kicad.new(ModuleLabel, instance)


class ModuleLine(HasLayerStrImpl, Selectable):
    """Wrapper for `EDGE_MODULE`"""
    @property
    def native_obj(self):
        return self._obj

    @staticmethod
    def wrap(instance):
        if type(instance) is not SWIGtype.FpShape:
            # raise TypeError()
            return None
        return kicad.new(ModuleLine, instance)


@deprecate_member('referenceLabel', 'reference_label')
@deprecate_member('valueLabel', 'value_label')
@deprecate_member('graphicalItems', 'graphical_items')
@deprecate_member('libName', 'lib_name')
@deprecate_member('fpName', 'fp_name')
class Module(HasPosition, HasRotation, Selectable, BoardItem):
    _ref_label = None
    _val_label = None

    def __init__(self, ref=None, pos=None, board=None):
        if not board:
            from kicad.pcbnew.board import Board
            try:
                board = Board.from_editor()
            except:
                board = None
        self._obj = SWIGtype.Footprint(board.native_obj)
        if ref:
            self.reference = ref
        if pos:
            self.position = pos
        if board:
            board.add(self)

    @staticmethod
    def wrap(instance):
        if type(instance) is not SWIGtype.Footprint:
            # raise TypeError()
            return None
        return kicad.new(Module, instance)

    @property
    def reference(self):
        return self._obj.GetReference()

    @reference.setter
    def reference(self, value):
        self._obj.SetReference(value)

    @property
    def reference_label(self):
        native = self._obj.Reference()
        if self._ref_label is None or self._ref_label.native_obj is not native:
            self._ref_label = ModuleLabel.wrap(native)
        return self._ref_label

    @property
    def value(self):
        return self._obj.GetValue()

    @value.setter
    def value(self, value):
        self._obj.SetValue(value)

    @property
    def value_label(self):
        native = self._obj.Value()
        if self._val_label is None or self._val_label.native_obj is not native:
            self._val_label = ModuleLabel.wrap(native)
        return self._val_label

    @property
    def graphical_items(self):
        """Text and drawings of module iterator."""
        for item in self._obj.GraphicalItems():
            if type(item) == SWIGtype.FpShape:
                yield ModuleLine.wrap(item)
            elif type(item) == SWIGtype.FpText:
                yield ModuleLabel.wrap(item)
            else:
                raise Exception("Unknown module item type: %s" % type(item))

    def flip(self):
        self._obj.Flip(self._obj.GetCenter())

    @property
    def layer(self):
        if self.board:
            return self.board.get_layer_name(self._obj.GetLayer())
        else:
            return pcbnew_layer.get_std_layer_name(self._obj.GetLayer())

    @layer.setter
    def layer(self, value):
        if value == self.layer:
            return
        if value not in ['F.Cu', 'B.Cu']:
            raise ValueError("You can place a module only on 'F.Cu' or 'B.Cu' layers!")
        # Using flip will make sure all components of the module are
        # switched to correct layer
        self.flip()

    @property
    def lib_name(self):
        return self._obj.GetFPID().GetLibNickname().GetChars()

    @property
    def fp_name(self):
        return self._obj.GetFPID().GetLibItemName().GetChars()

    def copy(self, ref, pos=None, board=None):
        """Create a copy of an existing module on the board
            A new reference designator is required, example:
                mod2 = mod1.copy('U2')
                mod2.reference == 'U2'  # True
        """
        if SWIG_version >= 7:
            _module = SWIGtype.Footprint(self._obj)
        else:
            _module = SWIGtype.Footprint(board and board._obj)
            _module.Copy(self._obj)
        module = Module.wrap(_module)
        module.reference = ref
        if pos:
            module.position = pos
        if board:
            board.add(module)
        elif self.board:
            self.board.add(module)
        return module

    @property
    def pads(self):
        for p in self._obj.Pads():
            yield Pad.wrap(p)

    def remove(self, element, permanent=False):
        ''' Makes it so Ctrl-Z works.
            Keeps a reference to the element in the python pcb object,
            so it persists for the life of that object
        '''
        if not permanent:
            if not hasattr(self, '_removed_elements'):
                self._removed_elements = []
            self._removed_elements.append(element)
        self._obj.Remove(element._obj)

    def restore_removed(self):
        if hasattr(self, '_removed_elements'):
            for element in self._removed_elements:
                self._obj.Add(element._obj)
        self._removed_elements = []


# In case v6+ naming is used
Footprint = Module
FootprintLine = ModuleLine
FootprintLabel = ModuleLabel



from kicad import pcbnew_bare as pcbnew

import kicad
from kicad.pcbnew import drawing
from kicad.pcbnew import module
from kicad.pcbnew.track import Track
from kicad.pcbnew.via import Via
from kicad.pcbnew.drawing import Drawing
from kicad.pcbnew.zone import Zone
from kicad import units, SWIGtype, instanceof


class _ModuleList(object):
    """Internal class to represent `Board.modules`"""
    def __init__(self, board):
        self._board = board

    def __getitem__(self, key):
        found = self._board._obj.FindFootprintByReference(key)
        if found:
            return module.Module.wrap(found)
        else:
            raise KeyError("No module with reference: %s" % key)

    def __iter__(self):
        # Note: this behavior is inconsistent with python manuals
        # suggestion. Manual suggests that a mapping should iterate
        # over keys ('reference' in our case). See:
        # https://docs.python.org/2.7/reference/datamodel.html?emulating-container-types#object.__iter__
        # But in my opinion `_ModuleList` is a list then mapping.
        for m in self._board._obj.GetFootprints():
            yield module.Module.wrap(m)

    def __len__(self):
        return len(self._board._obj.GetFootprints())

class Board(object):
    def __init__(self, wrap=None):
        """Board object"""
        if wrap:
            self._obj = wrap
        else:
            self._obj = pcbnew.BOARD()

        self._modulelist = _ModuleList(self)
        self._removed_elements = []

    @property
    def native_obj(self):
        return self._obj

    @staticmethod
    def wrap(instance):
        """Wraps a C++/old api BOARD object, and returns a Board."""
        return Board(wrap=instance)

    def add(self, obj):
        """Adds an object to the Board.

        Tracks, Drawings, Modules, etc...
        """
        self._obj.Add(obj.native_obj)
        return obj

    @property
    def modules(self):
        """Provides an iterator over the board Module objects."""
        return self._modulelist

    def moduleByRef(self, ref):
        """Returns the module that has the reference `ref`. Returns `None` if
        there is no such module."""
        found = self._obj.FindFootprintByReference(ref)
        if found:
            return module.Module.wrap(found)

    @property
    def footprints(self):
        """Alias footprint to module"""
        return self.modules

    def footprintByRef(self, ref):
        """Alias footprint to module"""
        return self.moduleByRef(ref)

    @property
    def vias(self):
        """An iterator over via objects"""
        for t in self._obj.GetTracks():
            if type(t) == SWIGtype.Via:
                yield Via.wrap(t)
            else:
                continue

    @property
    def tracks(self):
        """An iterator over track objects"""
        for t in self._obj.GetTracks():
            if type(t) == SWIGtype.Track:
                yield Track.wrap(t)
            else:
                continue

    @property
    def zones(self):
        """ An iterator over zone objects
            Implementation note: The iterator breaks if zones are removed during the iteration,
            so it is put in a list first, then yielded from that list.
            This issue was not seen with the other iterators
        """
        builder = list()
        for t in self._obj.Zones():
            if type(t) == SWIGtype.Zone:
                builder.append(Zone.wrap(t))
            else:
                continue
        for tt in builder:
            yield tt

    @property
    def drawings(self):
        all_drawings = []
        for drawing in self._obj.GetDrawings():
            if instanceof(drawing, (SWIGtype.Shape, SWIGtype.Text)):
                yield Drawing.wrap(drawing)

    @property
    def items(self):
        ''' Everything on the board '''
        for item in self.modules:
            yield item
        for item in self.vias:
            yield item
        for item in self.tracks:
            yield item
        for item in self.zones:
            yield item
        for item in self.drawings:
            yield item

    @staticmethod
    def from_editor():
        """Provides the board object from the editor."""
        return Board.wrap(pcbnew.GetBoard())

    @staticmethod
    def load(filename):
        """Loads a board file."""
        return Board.wrap(pcbnew.LoadBoard(filename))

    def save(self, filename=None):
        """Save the board to a file.

        filename should have .kicad_pcb extention.
        """
        if filename is None:
            filename = self._obj.GetFileName()
        self._obj.Save(filename)

    def copy(self):
        return Board(wrap=self._obj.Clone())

    # TODO: add setter for Board.filename
    @property
    def filename(self):
        """Name of the board file."""
        return self._obj.GetFileName()

    def add_module(self, ref, pos=(0, 0)):
        """Create new module on the board"""
        return module.Module(ref, pos, board=self)

    @property
    def default_width(self, width=None):
        b = self._obj
        return (
            float(b.GetDesignSettings().GetCurrentTrackWidth()) /
            units.DEFAULT_UNIT_IUS)

    def add_track_segment(self, start, end, layer='F.Cu', width=None):
        """Create a track segment."""

        track = Track(width or self.default_width,
                      start, end, layer, board=self)
        self._obj.Add(track.native_obj)
        return track

    def get_layer_id(self, name):
        lid = self._obj.GetLayerID(name)
        if lid == -1:
            # Try to recover from silkscreen rename
            if name == 'F.SilkS':
                lid = self._obj.GetLayerID('F.Silkscreen')
            elif name == 'F.Silkscreen':
                lid = self._obj.GetLayerID('F.SilkS')
            elif name == 'B.SilkS':
                lid = self._obj.GetLayerID('B.Silkscreen')
            elif name == 'B.Silkscreen':
                lid = self._obj.GetLayerID('B.Silkscreen')
        if lid == -1:
            raise ValueError('Layer {} not found in this board'.format(name))
        return lid

    def get_layer_name(self, layer_id):
        return self._obj.GetLayerName(layer_id)

    def add_track(self, coords, layer='F.Cu', width=None):
        """Create a track polyline.

        Create track segments from each coordinate to the next.
        """
        for n in range(len(coords) - 1):
            self.add_track_segment(coords[n], coords[n + 1],
                                   layer=layer, width=width)

    @property
    def default_via_size(self):
        return (float(self._obj.GetDesignSettings().GetCurrentViaSize()) /
                units.DEFAULT_UNIT_IUS)

    @property
    def default_via_drill(self):
        via_drill = self._obj.GetDesignSettings().GetCurrentViaDrill()
        if via_drill > 0:
            return (float(via_drill) / units.DEFAULT_UNIT_IUS)
        else:
            return 0.2

    def add_via(self, coord, layer_pair=('B.Cu', 'F.Cu'), size=None,
                drill=None):
        """Create a via on the board.

        :param coord: Position of the via.
        :param layer_pair: Tuple of the connected layers (for example
                           ('B.Cu', 'F.Cu')).
        :param size: size of via in mm, or None for current selection.
        :param drill: size of drill in mm, or None for current selection.
        :returns: the created Via
        """
        return self.add(
            Via(coord, layer_pair, size or self.default_via_size,
                drill or self.default_via_drill, board=self))

    def add_line(self, start, end, layer='F.SilkS', width=0.15):
        """Create a graphic line on the board"""
        return self.add(
            drawing.Segment(start, end, layer, width, board=self))

    def add_polyline(self, coords, layer='F.SilkS', width=0.15):
        """Create a graphic polyline on the board"""
        for n in range(len(coords) - 1):
            self.add_line(coords[n], coords[n + 1], layer=layer, width=width)

    def add_circle(self, center, radius, layer='F.SilkS', width=0.15):
        """Create a graphic circle on the board"""
        return self.add(
            drawing.Circle(center, radius, layer, width, board=self))

    def add_arc(self, center, radius, start_angle, stop_angle,
                layer='F.SilkS', width=0.15):
        """Create a graphic arc on the board"""
        return self.add(
            drawing.Arc(center, radius, start_angle, stop_angle,
                        layer, width, board=self))

    def add_text(self, position, text, layer='F.SilkS', size=1.0, thickness=0.15):
        return self.add(
            drawing.TextPCB(position, text, layer, size, thickness, board=self))

    def remove(self, element, permanent=False):
        ''' Makes it so Ctrl-Z works.
            Keeps a reference to the element in the python pcb object,
            so it persists for the life of that object
        '''
        if not permanent:
            self._removed_elements.append(element)
        self._obj.Remove(element._obj)

    def restore_removed(self):
        if hasattr(self, '_removed_elements'):
            for element in self._removed_elements:
                self._obj.Add(element._obj)
        self._removed_elements = []

    def deselect_all(self):
        self._obj.ClearSelected()

    @property
    def selected_items(self):
        ''' This useful for duck typing in the interactive terminal
            Suppose you want to set some drill radii. Iterating everything would cause attribute errors,
            so it is easier to just select the vias you want, then use this method for convenience.
            To get one item that you selected, use
                xx = next(pcb.selected_items)
        '''
        for item in self.items:
            try:
                if item.is_selected:
                    yield item
            except AttributeError:
                continue

    def fill_zones(self, zone_to_fill=None):
        ''' zone_to_fill=None fills all zones in this board '''
        filler = pcbnew.ZONE_FILLER(self._obj)
        filler.Fill(self._obj.Zones())



from kicad import pcbnew_bare as pcbnew

import kicad
from kicad import units
from kicad import Size
from kicad.exceptions import deprecate_member
from kicad.pcbnew.item import HasPosition, HasConnection, HasLayerStrImpl, Selectable, BoardItem

class DrillShape():
    Circle = pcbnew.PAD_DRILL_SHAPE_CIRCLE
    Oval = pcbnew.PAD_DRILL_SHAPE_OBLONG

class PadShape():
    Circle = pcbnew.PAD_SHAPE_CIRCLE
    Oval = pcbnew.PAD_SHAPE_OVAL
    Rectangle = pcbnew.PAD_SHAPE_RECT
    RoundedRectangle = pcbnew.PAD_SHAPE_ROUNDRECT
    Trapezoid = pcbnew.PAD_SHAPE_TRAPEZOID
    Chamfered = pcbnew.PAD_SHAPE_CHAMFERED_RECT
    Custom = pcbnew.PAD_SHAPE_CUSTOM

class PadType():
    Through = pcbnew.PAD_ATTRIB_PTH
    SMD = pcbnew.PAD_ATTRIB_SMD
    Connector = pcbnew.PAD_ATTRIB_CONN
    NPTH = pcbnew.PAD_ATTRIB_NPTH


@deprecate_member('padType', 'pad_type')
@deprecate_member('drillShape', 'drill_shape')
class Pad(HasPosition, HasConnection, HasLayerStrImpl, Selectable, BoardItem):
    def __init__(self):
        raise NotImplementedError('Direct instantiation of Pad is not supported')

    @staticmethod
    def wrap(instance):
        """Wraps a C++ api PAD object, and returns a `Pad`."""
        return kicad.new(Pad, instance)

    @property
    def pad_type(self):
        return self._obj.GetAttribute()

    @pad_type.setter
    def pad_type(self, value):
        """Value should be integer that can be found by referencing PadType.Through"""
        self._obj.SetAttribute(value)

    @property
    def drill_shape(self):
        return self._obj.GetDrillShape()

    @drill_shape.setter
    def drill_shape(self, value):
        """Value should be integer that can be found by referencing DrillShape.Circle"""
        self._obj.SetDrillShape(value)

    @property
    def drill(self):
        """Drill size. Returns `Size`."""
        return Size.wrap(self._obj.GetDrillSize())

    @drill.setter
    def drill(self, value):
        """Sets the drill size. If value is a single float or int, pad drill
        shape is set to circle, if input is a tuple of (x, y) drill
        shape is set to oval."""
        try:
            size = Size.build_from(value)
            self.drill_shape = DrillShape.Oval
        except TypeError:
            size = Size.build_from((value, value))
            self.drill_shape = DrillShape.Circle
        self._obj.SetDrillSize(size.native_obj)

    @property
    def shape(self):
        return self._obj.GetShape()

    @shape.setter
    def shape(self, value):
        """Value should be integer that can be found by referencing PadShape.Circle"""
        self._obj.SetShape(value)

    @property
    def size(self):
        return Size.wrap(self._obj.GetSize())

    @size.setter
    def size(self, value):
        try:
            size = Size.build_from(value)
            # self.drill_shape = DrillShape.Oval
        except TypeError:
            size = Size.build_from((value, value))
            # self.drill_shape = DrillShape.Circle
        self._obj.SetSize(size.native_obj)



import unittest

from kicad import *
from kicad.pcbnew.board import *


class TestPcbnewBoard(unittest.TestCase):
    #TODO(mangelajo): currently we're just testing that the parameter
    #                 combinations work, we may need to check that the
    #                 actual objects were created and added to the Board
    def setUp(self):
        self.board = Board()

    def test_module_creation(self):
        self.board.add_module('M1')
        self.assertEqual(1, len(list(self.board.modules)))
        self.board.add_module('M2')
        self.assertEqual(2, len(list(self.board.modules)))
        refs = [module.reference for module in self.board.modules]
        self.assertIn('M1', refs)
        self.assertIn('M2', refs)

    def test_track_segment_creation(self):
        self.board.add_track_segment((0, 0), (1, 1))
        self.board.add_track_segment((0, 0), (1, 1), layer='B.Cu')
        self.board.add_track_segment((0, 0), (1, 1), layer='B.Cu', width=2)

    def test_track_creation(self):
        self.board.add_track([(0, 0), (1, 1)])
        self.board.add_track([(1, 1), (1, 2)], 'B.Cu')
        self.board.add_track([(1, 2), (2, 2)], 'B.Cu', width=2)

    def test_via_creation(self):
        self.board.add_via((1, 1))
        self.board.add_via((1, 2), ('B.Cu', 'F.Cu'), size=2)
        self.board.add_via((1, 3), ('B.Cu', 'F.Cu'), size=2, drill=1)

    def test_default_via_props(self):
        self.assertGreater(self.board.default_via_size, 0.1)
        self.assertGreater(self.board.default_via_drill, 0.1)

    def test_line_creation(self):
        self.board.add_line((0, 0), (1, 1))
        self.board.add_line((0, 0), (1, 1), layer='B.SilkS')
        self.board.add_line((0, 0), (1, 1), layer='B.SilkS', width=1)

    def test_polyline_creation(self):
        self.board.add_polyline([(0, 0), (1, 1), (2, 2)])

    def test_add_circle(self):
        self.board.add_circle((1, 1), 1)

    def test_add_arc(self):
        self.board.add_arc((0, 0), 5, 0, 90)



import unittest

from kicad import *
from kicad.pcbnew.board import *


REFERENCE = 'M1'
OTHER_REFERENCE = 'M2'
POSITION = (1, 2)
OTHER_POSITION = (3, 4)


class TestPcbnewModule(unittest.TestCase):
    def setUp(self):
        self.board = Board()
        self.module = self.board.add_module(REFERENCE, pos=POSITION)

    def test_module_reference(self):
        self.assertEqual(REFERENCE, self.module.reference)
        self.module.reference = OTHER_REFERENCE
        self.assertEqual(OTHER_REFERENCE, self.module.reference)

    def test_module_position(self):
        self.assertEqual(POSITION, self.module.position)
        self.module.position = OTHER_POSITION
        self.assertEqual(OTHER_POSITION, self.module.position)

    def test_module_copy(self):
        module_copy = self.module.copy(OTHER_REFERENCE)
        self.assertEqual(OTHER_REFERENCE, module_copy.reference)
        self.assertEqual(POSITION, module_copy.position)

    def test_module_copy_in_board_and_position(self):
        module_copy = self.module.copy(
            OTHER_REFERENCE, pos=OTHER_POSITION, board=self.board)
        self.assertEqual(OTHER_POSITION, module_copy.position)



import unittest
from unittest.mock import patch
from functools import wraps
# from kicad.exceptions import deprecate_member
# import kicad.exceptions as kex  # will use deprecate_member and deprecate_warn_fun

deprecate_warn_fun = print  # print is sometimes good
def deprecate_member(old, new, deadline='v0.5.0'):
    def regular_decorator(klass):
        def auto_warn(fun):
            from_str = klass.__name__ + '.' + old
            to_str = klass.__name__ + '.' + new
            header = 'Deprecation warning (deadline {}):'.format(deadline)
            map_str = '{} -> {}'.format(from_str, to_str)
            @wraps(fun)
            def warner(*args, **kwargs):
                deprecate_warn_fun(header, map_str)
                return fun(*args, **kwargs)
            return warner

        new_meth = getattr(klass, new)
        if isinstance(new_meth, property):
            aug_meth = property(
                auto_warn(new_meth.fget),
                auto_warn(new_meth.fset),
                auto_warn(new_meth.fdel)
            )
        else:
            aug_meth = auto_warn(new_meth)
        setattr(klass, old, aug_meth)
        return klass
    return regular_decorator


@deprecate_member('myMeth', 'my_meth')
@deprecate_member('myClassmeth', 'my_classmeth')
@deprecate_member('myProp', 'my_prop')
class AugClass:
    def my_meth(self):
        ''' docstring here '''
        return 'my_meth'

    @classmethod
    def my_classmeth(cls):
        return 'my_classmeth'

    @property
    def my_prop(self):
        return 'my_prop.get'

    @my_prop.setter
    def my_prop(self, val):
        # return 'my_prop.set'
        pass

aug_obj = AugClass()
aug_obj.myProp

class MainTester(unittest.TestCase):
    @patch('__main__.deprecate_warn_fun')
    def test_1(self, mock_print):
        aug_obj = AugClass()

        self.assertEqual(aug_obj.my_meth(), 'my_meth')
        self.assertEqual(AugClass.my_classmeth(), 'my_classmeth')
        self.assertEqual(aug_obj.my_prop, 'my_prop.get')
        aug_obj.my_prop = 1
        mock_print.assert_not_called()

        self.assertEqual(aug_obj.myMeth(), 'my_meth')
        mock_print.assert_called_once()
        mock_print.reset_mock()
        self.assertEqual(AugClass.myClassmeth(), 'my_classmeth')
        mock_print.assert_called_once()
        mock_print.reset_mock()
        self.assertEqual(aug_obj.myProp, 'my_prop.get')
        mock_print.assert_called_once()
        mock_print.reset_mock()
        aug_obj.MyProp = 1
        # mock_print.assert_called_once()  # unittest misses, but tested manually to work

if __name__ == '__main__':
    unittest.main()



''' Automatic linker to pcbnew GUI and pcbnew python package
    Use this one time to create the link.

    Copy-paste the argument from pcbnew terminal.

    1. Run this in pcbnew's terminal window::

    # (OLD VERSION): print('link_kicad_python_to_pcbnew', pcbnew.__file__, pcbnew.GetKicadConfigPath())
    import pcbnew; print('link_kicad_python_to_pcbnew ' + pcbnew.__file__ + ' ' + pcbnew.SETTINGS_MANAGER.GetUserSettingsPath())


    Copy the output. It should look something like::

    link_kicad_python_to_pcbnew /usr/lib/python3/dist-packages/pcbnew.py /home/atait/.config/kicad

    2. Paste and run what you copied into command line from an environment where you have installed this package
'''
import os, sys
import argparse



# Tells this package where to find pcbnew module
pcbnew_path_store = os.path.join(os.path.dirname(__file__), '.path_to_pcbnew_module')


def get_pcbnew_path_from_file():
    if not os.path.isfile(pcbnew_path_store):
        return None
    with open(pcbnew_path_store) as fx:
        return fx.read().strip()


def get_pcbnew_path():
    # Look for the real pcbnew from environment and then file
    pcbnew_swig_path = os.environ.get('PCBNEW_PATH', get_pcbnew_path_from_file())
    if pcbnew_swig_path:
        # Validate
        if not os.path.basename(pcbnew_swig_path).startswith('pcbnew.py'):
            raise EnvironmentError(
                'Incorrect location for \'PCBNEW_PATH\' ({}).'
                ' It should point to a file called pcbnew.py'.format(pcbnew_swig_path))
        if not os.path.isfile(pcbnew_swig_path):
            raise EnvironmentError(
                'Incorrect location for \'PCBNEW_PATH\' ({}).'
                ' File does not exist'.format(pcbnew_swig_path))
    else:
        raise EnvironmentError(
            'You have not yet populated .path_to_pcbnew_module.'
            '\nUse the command "link_kicad_python_to_pcbnew" to set it up.')
    return pcbnew_swig_path


def get_pcbnew_module():
    try:
        return __import__('pcbnew')  # If this works, we are probably in the pcbnew application, and we're done.
    except ImportError:
        pass

    pcbnew_swig_path = get_pcbnew_path()
    if pcbnew_swig_path:
        # if 'Frameworks' in pcbnew_swig_path:
        #     dynlib_path = pcbnew_swig_path.split('Frameworks')[0] + 'Frameworks'
        #     sys.path.insert(0, dynlib_path)
        sys.path.insert(0, os.path.dirname(pcbnew_swig_path))
        try:
            pcbnew_bare = __import__('pcbnew')
        except ImportError as err:
            if err.args[0].startswith('dynamic module does not define'):
                print('You are likely using Mac or Windows,'
                      ' which means kicad does not yet support python 3 on your system.'
                      ' You will be able to use kicad-python within the pcbnew application,'
                      ' but not as a standalone.')
            else:
                print('Warning: pcbnew.py was located but could not be imported. It might be a python 2/3 issue:\n')
                print(err)
            pcbnew_bare = None
    else:
        # special case for documentation without pcbnew at all
        spoofing_for_documentation = os.environ.get('KICAD_PYTHON_IN_SPHINX_GENERATION', '0')
        if spoofing_for_documentation == '1':
            class SphinxEnumPhony:
                def __getattr__(self, attr):
                    return 0
            pcbnew_bare = SphinxEnumPhony()
        else:
            # failed to find pcbnew
            print(
                'pcbnew is required by kicad-python.'
                ' It gets installed when you install the kicad application, but not necessarily on your python path.'
                '\nSee instructions for how to link them at https://github.com/atait/kicad-python'
            )
            pcbnew_bare = None
    return pcbnew_bare


# Tells pcbnew application where to find this package
startup_script = """### Auto generated by kicad-python initialization for pcbnew console
import sys, pcbnew
sys.path.append("{}")
from kicad.pcbnew.board import Board
pcb = Board.from_editor()
"""

plugin_script = """### Auto generated by kicad-python initialization for pcbnew action plugins
import sys
sys.path.append("{}")
"""


def _print_file(arg):
    print('\033[4m\033[91m' + arg + '\033[0m')


def _print_contents(arg):
    print('\033[92m' + arg + '\033[0m')


def create_link(pcbnew_module_path, kicad_config_path, dry_run=False):
    writing = 'Would write' if dry_run else 'Writing'
    # Determine what to put in the startup script
    my_package_path = os.path.dirname(__file__)
    my_search_path = os.path.dirname(my_package_path)
    startup_contents = startup_script.format(my_search_path)
    # Determine where to put the startup script
    startup_file = os.path.join(kicad_config_path.strip(), 'PyShell_pcbnew_startup.py')

    # Check that we are not overwriting something
    write_is_safe = True
    if os.path.isfile(startup_file):
        with open(startup_file) as fx:
            line = fx.readline()
        if line.startswith('### DEFAULT STARTUP FILE'):
            pass
        elif line.startswith('### Auto generated by kicad-python'):
            pass
        else:
            write_is_safe = False

    # Write the startup script
    print('\n1. {} console startup script: for GUI snippet scripting'.format(writing))
    _print_file(startup_file)
    if write_is_safe:
        if not dry_run:
            with open(startup_file, 'w') as fx:
                fx.write(startup_contents)
        _print_contents(startup_contents)
    else:
        print('Warning: Startup file is not empty:\n', startup_file)
        print('It is safer to delete that file, or\ndo this manually by inserting these lines into that file:\n\n', startup_contents)

    # Write the plugin importer
    plugin_dir = os.path.join(kicad_config_path.strip(), 'scripting', 'plugins')
    os.makedirs(plugin_dir, exist_ok=True)
    plugin_file = os.path.join(plugin_dir, 'initialize_kicad_python_plugin.py')
    plugin_contents = plugin_script.format(my_search_path)
    print('2. {} plugin importer: for action plugin development'.format(writing))
    _print_file(plugin_file)
    if not dry_run:
        with open(plugin_file, 'w') as fx:
            fx.write(plugin_contents)
    _print_contents(plugin_contents)

    # Store the location of pcbnew module
    print('3. {} pcbnew path: for batch processing outside KiCAD'.format(writing))
    _print_file(pcbnew_path_store)
    if not dry_run:
        with open(pcbnew_path_store, 'w') as fx:
            fx.write(pcbnew_module_path.strip())
    _print_contents(pcbnew_module_path)

    # Try it
    if dry_run: return
    try:
        get_pcbnew_path()
    except ImportError as err:
        if err.args[0].startswith('dynamic module does not define'):
            print('You are likely using Mac or Windows,'
                  ' which means kicad does not yet support python 3 on your system.'
                  ' You will be able to use kicad-python in the pcbnew application,'
                  ' but not outside of it for batch processing.')
        else:
            raise
    else:
        print('Successfully linked kicad-python with pcbnew')


help_msg = """
Create bidirectional link between kicad-python and pcbnew.
To get the arguments correct, copy this and run it in pcbnew application console:

    >>> import pcbnew
    >>> pcbnew.__file__
    # [Path A]: This will give something like "/usr/lib/python3/dist-packages/pcbnew.py"
    >>> pcbnew.SETTINGS_MANAGER.GetUserSettingsPath()
    # [Path B]: This will give something like "home/username/.config/kicad"

For kicad 5, replace that last command with `pcbnew.SETTINGS_MANAGER_GetUserSettingsPath()` (note the last underscore).

Next, go back to your external command line or Terminal shell,
and run this command, replacing Path A and Path B with what you got above.

    link_kicad_python_to_pcbnew [Path A] [Path B]

For example,

    link_kicad_python_to_pcbnew /usr/lib/python3/dist-packages/pcbnew.py /home/username/.config/kicad
"""


parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter, description=help_msg)
parser.add_argument('pcbnew_module_path', type=str)
parser.add_argument('kicad_config_path', type=str)
parser.add_argument('-n', '--dry-run', action='store_true')

def cl_main():
    args = parser.parse_args()
    create_link(args.pcbnew_module_path, args.kicad_config_path, dry_run=args.dry_run)



from kicad import pcbnew_bare as pcbnew

import kicad
from kicad.pcbnew import layer as pcbnew_layer
from kicad.point import Point
from kicad import units, SWIGtype, SWIG_version
from kicad.pcbnew.item import HasPosition, HasConnection, Selectable, BoardItem


if SWIG_version >= 6:
    class ViaType():
        Through = pcbnew.VIATYPE_THROUGH
        Micro = pcbnew.VIATYPE_MICROVIA
        Blind = pcbnew.VIATYPE_BLIND_BURIED
else:
    class ViaType():
        Through = pcbnew.VIA_THROUGH
        Micro = pcbnew.VIA_MICROVIA
        Blind = pcbnew.VIA_BLIND_BURIED


class Via(HasPosition, HasConnection, Selectable, BoardItem):
    def __init__(self, coord, layer_pair, diameter, drill, board=None):
        self._obj = SWIGtype.Via(board and board.native_obj)
        self.diameter = diameter
        coord_point = Point.build_from(coord)
        self._obj.SetEnd(coord_point.native_obj)
        self._obj.SetStart(coord_point.native_obj)
        if board:
            self._obj.SetLayerPair(board.get_layer_id(layer_pair[0]),
                                   board.get_layer_id(layer_pair[1]))
        else:
            self._obj.SetLayerPair(pcbnew_layer.get_std_layer_id(layer_pair[0]),
                                   pcbnew_layer.get_std_layer_id(layer_pair[1]))
        self.drill = drill

    @staticmethod
    def wrap(instance):
        """Wraps a C++ api VIA object, and returns a `Via`."""
        return kicad.new(Via, instance)

    @property
    def drill(self):
        """Via drill diameter"""
        return float(self._obj.GetDrill()) / units.DEFAULT_UNIT_IUS

    @drill.setter
    def drill(self, value):
        self._obj.SetDrill(int(value * units.DEFAULT_UNIT_IUS))

    @property
    def diameter(self):
        """Via diameter"""
        return float(self._obj.GetWidth()) / units.DEFAULT_UNIT_IUS

    @diameter.setter
    def diameter(self, value):
        self._obj.SetWidth(int(value * units.DEFAULT_UNIT_IUS))

    @property
    def center(self):
        """Via center"""
        return Point.wrap(self._obj.GetCenter())

    @center.setter
    def center(self, value):
        self._obj.SetCenter(Point.native_from(value))

    @property
    def top_layer(self):
        if self.board:
            return self.board.get_layer_name(self._obj.TopLayer())
        else:
            return pcbnew_layer.get_std_layer_name(self._obj.TopLayer())

    @top_layer.setter
    def top_layer(self, value):
        if self.board:
            self._obj.SetTopLayer(self.board.get_layer_id(value))
        else:
            self._obj.SetTopLayer(pcbnew_layer.get_std_layer_id(value))

    @property
    def bottom_layer(self):
        if self.board:
            return self.board.get_layer_name(self._obj.BottomLayer())
        else:
            return pcbnew_layer.get_std_layer_name(self._obj.BottomLayer())

    @bottom_layer.setter
    def bottom_layer(self, value):
        if self.board:
            self._obj.SetTopLayer(self.board.get_layer_id(value))
        else:
            self._obj.SetTopLayer(pcbnew_layer.get_std_layer_name(value))

    @property
    def is_through(self):
        return self._obj.GetViaType() == ViaType.Through
        # self._obj.GetViaType() in [ViaType.Micro, ViaType.Blind]

    @is_through.setter
    def is_through(self, value):
        if value:
            self._obj.SetViaType(ViaType.Through)
        else:
            self._obj.SetViaType(ViaType.Blind)



from kicad import pcbnew_bare as pcbnew, SWIG_version
import kicad

class Layer():
    Front           = pcbnew.F_Cu
    Back            = pcbnew.B_Cu

    FrontAdhesive   = pcbnew.F_Adhes
    BackAdhesive    = pcbnew.B_Adhes
    FrontSilkScreen = pcbnew.F_SilkS
    BackSilkScreen  = pcbnew.B_SilkS
    FrontPaste      = pcbnew.F_Paste
    BackPaste       = pcbnew.B_Paste
    FrontMask       = pcbnew.F_Mask
    BackMask        = pcbnew.B_Mask

    DrawingsUser    = pcbnew.Dwgs_User
    CommentsUser    = pcbnew.Cmts_User
    ECO1User        = pcbnew.Eco1_User
    ECO2User        = pcbnew.Eco2_User

    EdgeCuts        = pcbnew.Edge_Cuts
    Margin          = pcbnew.Margin
    FrontFab        = pcbnew.F_Fab
    BackFab         = pcbnew.B_Fab
    FrontCourtyard  = pcbnew.F_CrtYd
    BackCourtyard   = pcbnew.B_CrtYd

# dicts for converting layer name to id, used by _get_layer
_std_layer_dict = None
_std_layer_names = None
def load_std_layers():
    # lazy import for Sphinx to run properly
    global _std_layer_dict, _std_layer_names
    if _std_layer_dict is None:
        if SWIG_version >= 7:
            native_get_layername = pcbnew.BOARD.GetStandardLayerName
        else:
            native_get_layername = pcbnew.BOARD_GetStandardLayerName
        _std_layer_dict = {native_get_layername(n): n
                           for n in range(pcbnew.PCB_LAYER_ID_COUNT)}
        try:
            # For backwards compatibility with silkscreen renames
            _std_layer_dict['F.SilkS'] = _std_layer_dict['F.Silkscreen']
            _std_layer_dict['B.SilkS'] = _std_layer_dict['B.Silkscreen']
        except KeyError:
            # Forwards compatibility
            _std_layer_dict['F.Silkscreen'] = _std_layer_dict['F.SilkS']
            _std_layer_dict['B.Silkscreen'] = _std_layer_dict['B.SilkS']
    if _std_layer_names is None:
        _std_layer_names = {s: n for n, s in _std_layer_dict.items()}


def get_board_layer_id(board, layer_name):
    """Get layer id for layer name in board, or std."""
    if board:
        return board.get_layer_id(layer_name)
    else:
        return get_std_layer_id(layer_name)


def get_board_layer_name(board, layer_id):
    """Get layer name for layer_id in board, or std."""
    if board:
        return board.get_layer_name(layer_id)
    else:
        return get_std_layer_name(layer_id)


def get_std_layer_name(layer_id):
    """Get layer name from layer id. """
    load_std_layers()
    return _std_layer_names[layer_id]


def get_std_layer_id(layer_name):
    """Get layer id from layer name

    If it is already an int just return it.
    """
    load_std_layers()
    return _std_layer_dict[layer_name]


class LayerSet:
    def __init__(self, layer_names, board=None):
        self._board = board
        self._build_layer_set(layer_names)

    @property
    def native_obj(self):
        return self._obj

    @staticmethod
    def wrap(instance):
        """Wraps a C++/old api LSET object, and returns a LayerSet."""
        return kicad.new(LayerSet, instance)

    def _build_layer_set(self, layers):
        """Create LayerSet used for defining pad layers"""
        bit_mask = 0
        for layer_name in layers:
            if self._board:
                bit_mask |= 1 << self._board.get_layer_id(layer_name)
            else:
                bit_mask |= 1 << get_std_layer_id(layer_name)
        hex_mask = '{0:013x}'.format(bit_mask)
        self._obj = pcbnew.LSET()
        self._obj.ParseHex(hex_mask, len(hex_mask))

    @property
    def layer_names(self):
        """Returns the list of layer names in this LayerSet."""
        return [get_board_layer_name(self._board, layer_id)
                for layer_id in self.layers]

    @property
    def layers(self):
        """Returns the list of Layer IDs in this LayerSet."""
        return [l for l in self._obj.Seq()]

    def add_layer(self, layer_name):
        self._obj.AddLayer(get_board_layer_id(self._board, layer_name))
        return self

    def remove_layer(self, layer_name):
        if layer_name not in self.layer_names:
            raise KeyError('Layer {} not present in {}'.format(layer_name, self.layer_names))
        self._obj.RemoveLayer(get_board_layer_id(self._board, layer_name))
        return self



from kicad import pcbnew_bare
import cmath

import kicad
from kicad import units
from kicad import SWIGtype


class Point(units.BaseUnitTuple):

    def __init__(self, x, y):
        """Creates a point.

        :param x: x coordinate.
        :param y: y coordinate.
        """
        self._class = Point
        self._obj = SWIGtype.Point(int(x * units.DEFAULT_UNIT_IUS),
                                   int(y * units.DEFAULT_UNIT_IUS))

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "Point(%g, %g)" % (self.x, self.y)

    @staticmethod
    def wrap(instance):
        """Wraps a wxPoint object from pcbnew and returns a Point one.

        This function should not be generally used, but it's provided as
        a helper when migrating old API code.

        :param instance: input wxPoint or VECTOR2I to wrap.
        :type instance: wxPoint
        :return: Point
        """
        wrapped_point = kicad.new(Point, instance)
        wrapped_point._class = Point
        return wrapped_point

    @staticmethod
    def build_from(t):
        """Return a point object from a tuple.

        It can transparently receive either a Point or a tuple,
        and a Point object will always be returned.
        """
        return Point._tuple_to_class(t, Point)

    @staticmethod
    def native_from(t):
        """Return a native C++/old API object from a tuple/Point.

        Generally not to be used, but provided for compatibility
        when migrating from old API code.
        """
        return Point._tuple_to_class(t, Point).native_obj

    @property
    def native_obj(self):
        """Returns the native wxPoint object Point is wrapping."""
        return self._obj

    def rotate(self, angle, around=(0, 0)):
        """Rotate the point.

        :param angle: rotation angle in degrees.
        :param around: rotation center.
        """
        self.x, self.y = self._rotated(angle, around)

    def rotated(self, angle, around=(0, 0)):
        """Generate a new Point.

        :param angle: rotation angle in degrees.
        :param around: rotation center.
        :returns: Point
        """
        x, y = self._rotated(angle, around)
        return Point(x, y)

    def _rotated(self, angle, around=(0, 0)):
        """Rotate coordinate around another point"""
        around = Point.build_from(around)
        p0 = self - around
        coord = (p0.x + p0.y * 1j) * cmath.exp((angle / units.rad) * 1j)
        return (coord.real + around.x, coord.imag + around.y)



__version__ = '0.3.9'

#: centralized import with fallback.
#: Necessary for documentation and environment patching outside of application
#: Import in this module in subpackages with
#: from kicad import pcbnew_bare as pcbnew
import os, sys
from kicad.environment import get_pcbnew_module
from kicad.exceptions import notify

# Find SWIG pcbnew
try:
    pcbnew_bare = get_pcbnew_module()
except EnvironmentError:
    print('Warning: pcbnew.py is not found or PCBNEW_PATH is corrupted. '
        'Only environment commands will be available')
    pcbnew_bare = None


# Low-level "new" function that avoids initializer
class BareClass(object):
    pass


def new(class_type, instance):
    """Returns an object of class without calling __init__.

    This could lead to inconsistent objects, use only when you
    know what you're doing.
    In kicad-python this is used to construct wrapper classes
    before injecting the native object.
    """
    obj = BareClass()
    obj.__class__ = class_type
    obj._obj = instance
    return obj


# Unify v5/6/7 APIs
if pcbnew_bare is None:
    SWIG_version = None
    class SWIGtype: pass
else:
    # Determine version and map equivalent objects into consistent names
    try:
        ver = tuple(int(x) for x in pcbnew_bare.GetMajorMinorVersion().split('.'))
    except AttributeError:
        ver = (5, 0)
    if ver[0] == 7 or (ver[0] == 6 and ver[1] == 99):
        SWIG_version = 7
    elif ver[0] == 6 or (ver[0] == 5 and ver[1] == 99):
        SWIG_version = 6
    elif ver[0] == 5 or (ver[0] == 4 and ver[1] == 99):
        SWIG_version = 5
    else:
        print('Version {} not supported by kicad-python. Some functionality might not work')
        SWIG_version = 7 if ver[0] > 7 else 5

    if SWIG_version == 7:
        class SWIGtype:
            Zone = pcbnew_bare.ZONE
            Track = pcbnew_bare.PCB_TRACK
            Via = pcbnew_bare.PCB_VIA
            Shape = pcbnew_bare.PCB_SHAPE
            Text = pcbnew_bare.PCB_TEXT
            Footprint = pcbnew_bare.FOOTPRINT
            FpText = pcbnew_bare.FP_TEXT
            FpShape = pcbnew_bare.FP_SHAPE
            # Changed in v7
            Point = pcbnew_bare.VECTOR2I
            Size = pcbnew_bare.VECTOR2I
            Rect = pcbnew_bare.BOX2I
            # End v7 changes
    elif SWIG_version == 6:
        class SWIGtype:
            # Changed in v6
            Zone = pcbnew_bare.ZONE
            Track = pcbnew_bare.PCB_TRACK
            Via = pcbnew_bare.PCB_VIA
            Shape = pcbnew_bare.PCB_SHAPE
            Text = pcbnew_bare.PCB_TEXT
            Footprint = pcbnew_bare.FOOTPRINT
            FpText = pcbnew_bare.FP_TEXT
            FpShape = pcbnew_bare.FP_SHAPE
            # End v6 changes
            Point = pcbnew_bare.wxPoint
            Size = pcbnew_bare.wxSize
            Rect = pcbnew_bare.EDA_RECT
    elif SWIG_version == 5:
        class SWIGtype:
            Zone = pcbnew_bare.ZONE_CONTAINER
            Track = pcbnew_bare.TRACK
            Via = pcbnew_bare.VIA
            Shape = pcbnew_bare.DRAWSEGMENT
            Text = pcbnew_bare.TEXTE_PCB
            Footprint = pcbnew_bare.MODULE
            FpText = pcbnew_bare.TEXTE_MODULE
            FpShape = pcbnew_bare.EDGE_MODULE
            Point = pcbnew_bare.wxPoint
            Size = pcbnew_bare.wxSize
            Rect = pcbnew_bare.EDA_RECT


# Broken isinstance detection of inheritance in v7
def instanceof(item, klass):
    if isinstance(klass, (tuple, list)):
        for kls in klass:
            if instanceof(item, kls):
                return True
    if isinstance(item, klass):  # This should hit in v6
        return True
    class_of_name = klass.__name__ + '_ClassOf'
    try:
        class_of_fun = getattr(pcbnew_bare, class_of_name)
        return class_of_fun(item)
    except AttributeError:
        return False

# Get reload: useful for on-the-fly updates to action plugin scripts without refreshing plugins
try:
    from importlib import reload
except ImportError:
    try:
        from imp import reload
    except ImportError:
        try:
            _ = reload
        except NameError as err:
            def reload(mod):
                pass


# Expose the basic classes to this package's top level
if pcbnew_bare:
    from .units import *
    from .point import Point
    from .size import Size
    import kicad.pcbnew



import cmath
import math

from kicad import pcbnew_bare as pcbnew, instanceof
import kicad
from kicad.pcbnew import layer as pcbnew_layer
from kicad.point import Point
from kicad import units, Size, SWIGtype, SWIG_version
from kicad.pcbnew.item import HasLayerStrImpl, Selectable, HasPosition, HasWidth, BoardItem

class ShapeType():
    Segment = pcbnew.S_SEGMENT
    Circle = pcbnew.S_CIRCLE
    Arc = pcbnew.S_ARC
    Polygon = pcbnew.S_POLYGON
    Rect = pcbnew.S_RECT

class Drawing(HasLayerStrImpl, HasPosition, HasWidth, Selectable, BoardItem):
    @staticmethod
    def wrap(instance):
        if instanceof(instance, SWIGtype.Shape):
            return Drawing._wrap_drawsegment(instance)
        elif instanceof(instance, SWIGtype.Text):
            return kicad.new(TextPCB, instance)
        else:
            raise TypeError('Invalid drawing class: {}'.format(type(instance)))

    @staticmethod
    def _wrap_drawsegment(instance):
        obj_shape = instance.GetShape()

        if obj_shape is pcbnew.S_SEGMENT:
            return kicad.new(Segment, instance)

        if obj_shape is pcbnew.S_CIRCLE:
            return kicad.new(Circle, instance)

        if obj_shape is pcbnew.S_ARC:
            return kicad.new(Arc, instance)

        if obj_shape is pcbnew.S_POLYGON:
            return kicad.new(Polygon, instance)

        if obj_shape is pcbnew.S_RECT:
            return kicad.new(Rectangle, instance)

        # Time to fail
        layer = instance.GetLayer()
        layer_str = pcbnew.BOARD_GetStandardLayerName(layer)
        unsupported = ['S_CURVE', 'S_LAST']
        for unsup in unsupported:
            if not hasattr(pcbnew, unsup):
                continue
            if obj_shape is getattr(pcbnew, unsup):
                raise TypeError('Unsupported shape type: pcbnew.{} on layer {}.'.format(unsup, layer_str))

        raise TypeError('Unrecognized shape type on layer {}'.format(layer_str))


class Segment(Drawing):
    def __init__(self, start, end, layer='F.SilkS', width=0.15, board=None):
        line = SWIGtype.Shape(board and board.native_obj)
        line.SetShape(ShapeType.Segment)
        self._obj = line
        line.SetStart(Point.native_from(start))
        line.SetEnd(Point.native_from(end))
        self.layer = layer
        self.width = width

    @property
    def start(self):
        return Point.wrap(self._obj.GetStart())

    @start.setter
    def start(self, value):
        self._obj.SetStart(Point.native_from(value))

    @property
    def end(self):
        return Point.wrap(self._obj.GetEnd())

    @end.setter
    def end(self, value):
        self._obj.SetEnd(Point.native_from(value))


class Circle(Drawing):
    def __init__(self, center, radius, layer='F.SilkS', width=0.15,
                 board=None):
        circle = SWIGtype.Shape(board and board.native_obj)
        circle.SetShape(ShapeType.Circle)
        circle.SetCenter(Point.native_from(center))
        start_coord = Point.native_from(
            (center[0], center[1] + radius))
        if SWIG_version >= 6:
            circle.SetEnd(start_coord)
            circle.SetModified()
        else:
            circle.SetArcStart(start_coord)
        circle.SetLayer(pcbnew_layer.get_board_layer_id(board, layer))
        circle.SetWidth(int(width * units.DEFAULT_UNIT_IUS))
        self._obj = circle

    @property
    def center(self):
        return Point.wrap(self._obj.GetCenter())

    @center.setter
    def center(self, value):
        self._obj.SetCenter(Point.native_from(value))

    @property
    def start(self):
        if SWIG_version >= 6:
            return Point.wrap(self._obj.GetEnd())
        else:
            return Point.wrap(self._obj.GetArcStart())

    @start.setter
    def start(self, value):
        if SWIG_version >= 6:
            self._obj.SetEnd(Point.native_from(value))
            self._obj.SetModified()
        else:
            self._obj.SetArcStart(Point.native_from(value))

    @property
    def radius(self):
        return float(self._obj.GetRadius()) / units.DEFAULT_UNIT_IUS

    @radius.setter
    def radius(self, value):
        self._obj.SetRadius(int(value * units.DEFAULT_UNIT_IUS))


# --- Logic for Arc changed a lot in version 6, so there are two classes
class Arc_v5(Drawing):
    def __init__(self, center, radius, start_angle, stop_angle,
                 layer='F.SilkS', width=0.15, board=None):
        start_coord = radius * cmath.exp(math.radians(start_angle - 90) * 1j)
        start_coord = Point.native_from((start_coord.real, start_coord.imag))
        center_coord = Point.native_from(center)
        start_coord += center_coord

        angle = stop_angle - start_angle
        arc = SWIGtype.Shape(board and board.native_obj)
        arc.SetShape(ShapeType.Arc)
        arc.SetCenter(center_coord)
        arc.SetArcStart(start_coord)
        arc.SetAngle(angle * 10)
        arc.SetLayer(pcbnew_layer.get_board_layer_id(board, layer))
        arc.SetWidth(int(width * units.DEFAULT_UNIT_IUS))
        self._obj = arc

    @property
    def center(self):
        return Point.wrap(self._obj.GetCenter())

    @center.setter
    def center(self, value):
        self._obj.SetCenter(Point.native_from(value))

    @property
    def radius(self):
        return float(self._obj.GetRadius()) / units.DEFAULT_UNIT_IUS

    @radius.setter
    def radius(self, value):
        self._obj.SetRadius(int(value * units.DEFAULT_UNIT_IUS))

    @property
    def start(self):
        return Point.wrap(self._obj.GetArcStart())

    @start.setter
    def start(self, value):
        self._obj.SetArcStart(Point.native_from(value))

    @property
    def end(self):
        return Point.wrap(self._obj.GetArcEnd())

    @end.setter
    def end(self, value):
        self._obj.SetArcEnd(Point.native_from(value))

    @property
    def angle(self):
        return float(self._obj.GetAngle()) / 10

    @angle.setter
    def angle(self, value):
        self._obj.SetAngle(value * 10)


class Arc_v6(Drawing):
    def __init__(self, center, radius, start_angle, stop_angle,
                 layer='F.SilkS', width=0.15, board=None):
        start_coord = radius * cmath.exp(math.radians(start_angle - 90) * 1j)
        abs_start = (start_coord.real + center[0], start_coord.imag + center[1])

        arc = SWIGtype.Shape(board and board.native_obj)
        arc.SetShape(ShapeType.Arc)
        self._obj = arc
        self.center = center
        self.start = abs_start
        self.angle = stop_angle - start_angle
        self.layer = layer
        self.width = width

    @property
    def center(self):
        return Point.wrap(self._obj.GetCenter())

    @center.setter
    def center(self, value):
        self._obj.SetCenter(Point.native_from(value))

    @property
    def radius(self):
        return float(self._obj.GetRadius()) / units.DEFAULT_UNIT_IUS

    @radius.setter
    def radius(self, value):
        self._obj.SetRadius(int(value * units.DEFAULT_UNIT_IUS))

    @property
    def start(self):
        return Point.wrap(self._obj.GetStart())

    @start.setter
    def start(self, value):
        self._obj.SetStart(Point.native_from(value))

    @property
    def end(self):
        return Point.wrap(self._obj.GetEnd())

    @end.setter
    def end(self, value):
        start = self._obj.GetStart()
        mid = self._obj.GetArcMid()
        self._obj.SetArcGeometry(start, mid, Point.native_from(value))

    @property
    def angle(self):
        if SWIG_version >= 7:
            return float(self._obj.GetArcAngle().AsDegrees())
        else:
            return float(self._obj.GetArcAngle()) / 10

    @angle.setter
    def angle(self, value):
        if SWIG_version >= 7:
            val_obj = pcbnew.EDA_ANGLE(value, pcbnew.EDA_UNITS_DEGREES)
            self._obj.SetArcAngleAndEnd(val_obj)
        else:
            self._obj.SetArcAngleAndEnd(value * 10)

if SWIG_version >= 6:
    Arc = Arc_v6
else:
    Arc = Arc_v5


class Polygon(Drawing):
    def __init__(self, coords,
                 layer='F.SilkS', width=0.15, board=None):
        poly_obj = SWIGtype.Shape(board and board.native_obj)
        poly_obj.SetShape(ShapeType.Polygon)
        self._obj = poly_obj

        chain = pcbnew.SHAPE_LINE_CHAIN()
        for coord in coords:
            chain.Append(Point.native_from(coord))
        chain.SetClosed(True)
        poly_shape = pcbnew.SHAPE_POLY_SET(chain)
        poly_obj.SetPolyShape(poly_shape)

        self.layer = layer
        self.width = width

    @property
    def filled(self):
        return self._obj.IsFilled()

    @filled.setter
    def filled(self, value=True):
        self._obj.SetFilled(value)

    def get_vertices(self):
        poly = self._obj.GetPolyShape()
        noutlines = poly.OutlineCount()
        if noutlines == 0:
            raise RuntimeError('Polygon\'s SHAPE_POLY_SET has no Outlines')
        elif noutlines > 1:
            raise ValueError('Polygon contains multiple Outlines which is not supported')
        outline = poly.Outline(0)
        pts = []
        for ipt in range(outline.PointCount()):
            native = outline.GetPoint(ipt)
            pts.append(Point.wrap(native))
        return pts

    def to_segments(self, replace=False):
        ''' If replace is true, removes the original polygon
        '''
        segs = []
        verts = self.get_vertices()
        for iseg in range(len(verts)):
            new_seg = Segment(verts[iseg-1], verts[iseg],
                self.layer, self.width, self.board
            )
            segs.append(new_seg)
        if replace:
            for seg in segs:
                self.board.add(seg)
            self.board.remove(self)
        return segs

    def fillet(self, radius_mm, tol_mm=.01):
        poly = self.native_obj.GetPolyShape()
        smoothed = poly.Fillet(int(radius_mm * units.DEFAULT_UNIT_IUS), int(tol_mm * units.DEFAULT_UNIT_IUS))
        self.native_obj.SetPolyShape(smoothed)

    def contains(self, point):
        poly = self._obj.GetPolyShape()
        return poly.Contains(Point.native_from(point))


class Rectangle(Polygon):
    ''' Inherits x,y get/set from HasPosition '''
    def __init__(self, corner_nw, corner_se,
                 layer='F.SilkS', width=0.15, board=None):
        rect_obj = SWIGtype.Shape(board and board.native_obj)
        rect_obj.SetShape(ShapeType.Rect)
        self._obj = rect_obj
        rect_obj.SetStart(Point.native_from(corner_nw))
        rect_obj.SetEnd(Point.native_from(corner_se))
        self.layer = layer
        self.width = width

    @classmethod
    def from_centersize(cls, xcent, ycent, xsize, ysize,
                     layer='F.SilkS', width=0.15, board=None):
        center = Point(xcent, ycent)
        half_size = Point(xsize / 2, ysize / 2)
        corner_nw = center - half_size
        corner_se = center + half_size
        return cls(corner_nw, corner_se, layer, width, board)

    def get_vertices(self):
        corners_native = self.native_obj.GetRectCorners()
        corners = [Point.wrap(pt) for pt in corners_native]
        return corners

    @property
    def size(self):
        nw = Point.wrap(self._obj.GetStart())
        se = Point.wrap(self._obj.GetEnd())
        sz = nw - se
        return (abs(sz[0]), abs(sz[1]))

    # The inherited to_segments works based on overloading get_vertices

    def to_polygon(self, replace=False):
        corners_native = self.native_obj.GetRectCorners()
        corners = [Point.wrap(pt) for pt in corners_native]
        poly = Polygon(corners, layer=self.layer, width=self.width, board=self.board)
        if replace:
            self.board.add(poly)
            self.board.remove(self)
        return poly

    def fillet(self, radius_mm, tol_mm=.01):
        ''' Deletes the rectangle but that is ok in most situations
            It can be undone IF it is run inside an action plugin
        '''
        poly = self.to_polygon(replace=True)
        poly.fillet(radius_mm, tol_mm)

    def contains(self, point):
        poly = self.to_polygon(replace=False)
        return poly.contains(point)


class TextPCB(Drawing):
    def __init__(self, position, text=None, layer='F.SilkS',
                 size=1.0, thickness=0.15, board=None):
        self._obj = SWIGtype.Text(board and board.native_obj)
        self.position = position
        if text:
            self.text = text
        self.layer = layer
        self.size = size
        self.thickness = thickness

    @property
    def text(self):
        return self._obj.GetText()

    @text.setter
    def text(self, value):
        return self._obj.SetText(value)

    @property
    def thickness(self):
        if SWIG_version >= 7:
            return float(self._obj.GetTextThickness()) / units.DEFAULT_UNIT_IUS
        else:
            return float(self._obj.GetThickness()) / units.DEFAULT_UNIT_IUS

    @thickness.setter
    def thickness(self, value):
        if SWIG_version >= 7:
            return self._obj.SetTextThickness(int(value * units.DEFAULT_UNIT_IUS))
        else:
            return self._obj.SetThickness(int(value * units.DEFAULT_UNIT_IUS))

    @property
    def size(self):
        return Size.wrap(self._obj.GetTextSize())

    @size.setter
    def size(self, value):
        try:
            size = Size.build_from(value)
        except TypeError:
            size = Size.build_from((value, value))
        self._obj.SetTextSize(size.native_obj)

    @property
    def orientation(self):
        return self._obj.GetTextAngle() / 10

    @orientation.setter
    def orientation(self, value):
        self._obj.SetTextAngle(value * 10)

    @property
    def justification(self):
        hj = self._obj.GetHorizJustify()
        vj = self._obj.GetVertJustify()
        for k, v in justification_lookups.items():
            if hj == getattr(pcbnew, v):
                hjs = k
            if vj in getattr(pcbnew, v):
                vjs = k
        return hjs, vjs

    @justification.setter
    def justification(self, value):
        if isinstance(value, (list, tuple)):
            assert len(value) == 2
            self.justification = value[0]
            self.justification = value[1]
        else:
            try:
                token = justification_lookups[value]
            except KeyError:
                raise ValueError('Invalid justification {} of available {}'.format(value, list(justification_lookups.keys())))
            enum_val = getattr(pcbnew, token)
            if 'HJUSTIFY' in token:
                self._obj.SetHorizJustify(enum_val)
            else:
                self._obj.SetVertJustify(enum_val)

justification_lookups = dict(
    left='GR_TEXT_HJUSTIFY_LEFT',
    center='GR_TEXT_HJUSTIFY_CENTER',
    right='GR_TEXT_HJUSTIFY_RIGHT',
    bottom='GR_TEXT_VJUSTIFY_BOTTOM',
    middle='GR_TEXT_VJUSTIFY_CENTER',
    top='GR_TEXT_VJUSTIFY_TOP',
)



import unittest

from kicad import *


class TestKicadPoint(unittest.TestCase):

    def test_rotation(self):
        p = Point(1, 1)
        x, y = p.rotated(180).mm
        self.assertAlmostEqual(x, -1.0, places=5)
        self.assertAlmostEqual(y, -1.0, places=5)

    def test_rotation_around(self):
        p = Point(1, 1)
        p_rotated = p.rotated(180, around=(1, 1))
        self.assertEqual(p_rotated.mm, (1.0, 1.0))

    def test_equal(self):
        p1 = Point(2, 2)
        p2 = Point(2, 2)
        self.assertTrue(p1 == p2)
        self.assertTrue(p1 == (2, 2))



import unittest

from kicad.pcbnew.layer import *


TEST_LAYER_LIST = ['F.Cu', 'B.Cu', ]


class TestPcbnewLayerSet(unittest.TestCase):
    def test_layer_names(self):
        layer_set = LayerSet(TEST_LAYER_LIST)
        self.assertEqual(sorted(layer_set.layer_names),
                         sorted(TEST_LAYER_LIST))

    def test_layer_ids(self):
        layer_set = LayerSet(TEST_LAYER_LIST)
        self.assertEqual(len(layer_set.layers), len(TEST_LAYER_LIST))



#!/usr/bin/env python
import setuptools

try:
    reqs = open('requirements.txt', 'r').read().splitlines()
except IOError:
    reqs = []

def readme():
    with open('README.md', 'r') as fx:
      return fx.read()

setuptools.setup(
        name='kigadgets',
        version='0.3.9',
        description='Cross-version python wrapper for KiCAD pcbnew',
        long_description=readme(),
        long_description_content_type='text/markdown',
        author='Piers Titus van der Torren, Miguel Angel Ajo, Hasan Yavuz Ozderya, Alex Tait',
        author_email='atait@ieee.org',
        license='GPL v2',
        url='https://github.com/atait/kicad-python/',
        packages=setuptools.find_packages(exclude=['ez_setup']),
        install_requires=reqs,
        entry_points={'console_scripts': 'link_kicad_python_to_pcbnew=kicad.environment:cl_main'},
    )



import math

nm = 0.000001
mm = 1.0
inch = 25.4
mil = 0.25400

nm_ius = 1
mm_ius = 1000000
inch_ius = 254000000
mil_ius = 25400

deg = 1.0
rad = 180.0 / math.pi

DEFAULT_UNIT_IUS = mm_ius


def inch_to_mm(val):
    """Convert from inch to mm.

    Handles single values, sequences, sequences of sequences, etc.
    """
    try:
        return val * 25.4
    except TypeError:
        return [inch_to_mm(v) for v in val]


def mm_to_inch(val):
    """Convert from mm to inch.

    Handles single values, sequences, sequences of sequences, etc.
    """
    try:
        return val / 25.4
    except TypeError:
        return [mm_to_inch(v) for v in val]


class BaseUnitTuple(object):
    """Base class to provide mm, inch, mil properties.

    It's a class to be used just by Point and Size.
    """
    @property
    def x(self):
        """x coordinate."""
        return float(self.native_obj.x) / DEFAULT_UNIT_IUS

    @x.setter
    def x(self, value):
        self.native_obj.x = int(value * DEFAULT_UNIT_IUS)

    @property
    def y(self):
        """y coordinate."""
        return float(self.native_obj.y) / DEFAULT_UNIT_IUS

    @y.setter
    def y(self, value):
        self.native_obj.y = int(value * DEFAULT_UNIT_IUS)

    def __getitem__(self, index):
        return self.mm[index]

    def __setitem__(self, index, value):
        if index == 0:
            self.x = value
        elif index == 1:
            self.y = value
        else:
            raise IndexError

    def __sub__(self, b):
        return self._class.build_from((self[0] - b[0],
                                       self[1] - b[1]))

    def __add__(self, b):
        return self._class.build_from((self[0] + b[0],
                                       self[1] + b[1]))

    def __eq__(self, _other):
        other = self._class.build_from(_other)
        return (self.native_obj == other.native_obj)

    def __ne__(self, other):
        return not (self == other)

    def __len__(self):
        return len(self.native_obj)

    def _unit_tuple(self, unit_multiplier):
        unit_multiplier_float = float(unit_multiplier)
        return (self.x / unit_multiplier_float,
                self.y / unit_multiplier_float)

    @property
    def nm(self):
        """Get the nanometers tuple."""
        return self._unit_tuple(nm)

    @property
    def mm(self):
        """Get the milimeters tuple.

        :Example:

        >>> import kicad
        >>> p = kicad.Point(1, 2)
        >>> p.mm
        (1, 2)

        """
        return (self.x, self.y)

    @property
    def inch(self):
        """Get the inches tuple."""
        return self._unit_tuple(inch)

    @property
    def mil(self):
        """Get the mils tuple."""
        return self._unit_tuple(mil)

    @staticmethod
    def _tuple_to_class(v, cls):
        if isinstance(v, cls):
            return v
        elif isinstance(v, (tuple, list)):
            if len(v) != 2:
                raise TypeError("A point parameter must be a 2 value tuple")
        return cls(v[0], v[1])
